kskuForageExtension = {};

function kskuForageExtension.prerequisitesPresent(specializations)
    return true;
end;

AIDriveStrategyCombine.getDriveData = Utils.overwrittenFunction(AIDriveStrategyCombine.getDriveData, function (self, superFunc, dt, vX, vY, vZ)
    local isTurning = false;
    local allowedToDrive = true;

    for _,combine in pairs(self.combines) do
        for _, implement in pairs(combine.spec_attacherJoints:getAttachedImplements()) do
            combine.implement = implement.object;
            break;
        end;
        if combine.implement.needToStop and not isTurning then
            allowedToDrive = false;
        else
            allowedToDrive = true;
        end;
    end;

    if not allowedToDrive then
        return 0, 1, true, 0, math.huge;
    end;

    if superFunc ~= nil then
        return superFunc(self, dt, vX,vY,vZ);
    end;
    return nil, nil, nil, nil, nil;
end)

function kskuForageExtension.registerFunctions(vehicleType)
    local newFunctions = {
        "lockCombine",
        "resetTimer",
        "startBeaconLights",
        "stopBeaconLights",
        "activateChaffSwath",
        "deactivateChaffSwath"
    };

    for _, newFunction in ipairs(newFunctions) do
        SpecializationUtil.registerFunction(vehicleType, newFunction, kskuForageExtension[newFunction]);
    end;
end;

function kskuForageExtension.registerEventListeners(vehicleType)
    local functionNames = {
        "onLoad",
        "onUpdate",
        "onRegisterActionEvents"
    };

    for _, functionName in ipairs(functionNames) do
        SpecializationUtil.registerEventListener(vehicleType, functionName, kskuForageExtension);
    end;
end;

function kskuForageExtension:onRegisterActionEvents(isSelected, isOnActiveVehicle)
    local actionEventId;
    if isOnActiveVehicle then
        _, actionEventId = InputBinding.registerActionEvent(g_inputBinding, 'EMPTY_CHAFF_BUFFER', self, kskuForageExtension.actionEventEmptyChaffBuffer, false, true, false, true, nil);
        g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
        g_inputBinding:setActionEventTextVisibility(actionEventId, true);
    end;
end;

function kskuForageExtension:actionEventEmptyChaffBuffer(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_forageWagon;
    local currentFillType = self:getFillUnitFillType(spec.fillUnitIndex);

    if currentFillType ~= FillType.UNKNOWN then
        self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, -math.huge, currentFillType, ToolType.UNDEFINED, nil);
    end;
end;

function kskuForageExtension:onLoad(savegame)
    local specForageWagon = self.spec_forageWagon;

    self:resetTimer(specForageWagon);

    specForageWagon.stopTimerMax = 2000; --## stop combine for 2 seconds
    specForageWagon.stopCPTimerMax = 2000; --## stop combine for 2 seconds
    specForageWagon.needTurnOnBeaconLights = true;
    specForageWagon.beaconLightsWasTurnedOn = false;
    specForageWagon.needTurnOffStrawSwath = true;
    specForageWagon.chaffSwathWasTurnedOff = false;
    specForageWagon.mapHotspotBlinkingActive = false;
    specForageWagon.notificationGrainTankWarningShown = false;
end;

function kskuForageExtension:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self:getIsActive() then
        local modCP = g_modIsLoaded["FS22_Courseplay"];
        local specDischargeable = self.spec_dischargeable.currentDischargeNode;
        local specForageWagon = self.spec_forageWagon;
        local specDrivable = self.spec_drivable;
        local isMotorStarted = self.getIsMotorStarted ~= nil and self:getIsMotorStarted();
        local controlledTractor = nil;
        local vehicleToCheck = self.getRootVehicle and self:getRootVehicle();

        if self.getAttacherVehicle ~= nil then
            controlledTractor = self:getAttacherVehicle();
            controlledCombine = controlledTractor.spec_combine;
        end;

        if controlledTractor ~= nil then
            specDrivable = controlledTractor.spec_drivable;
            isMotorStarted = controlledTractor.getIsMotorStarted ~= nil and controlledTractor:getIsMotorStarted();
        end;

        if specDrivable ~= nil and controlledTractor ~= nil and isMotorStarted then
            local actionEvent = controlledCombine.actionEvents[InputAction.TOGGLE_CHOPPER];
            if actionEvent ~= nil then
                if controlledCombine.isSwathActive then
                    text = g_i18n:getText("action_disableChaffSwath");
                else
                    text = g_i18n:getText("action_enableChaffSwath");
                end;

                g_inputBinding:setActionEventText(actionEvent.actionEventId, text);
            end;

            local currentPipeTargetState = specForageWagon.spec_pipe.targetState;
            local trailer = NetworkUtil.getObject(specForageWagon.spec_pipe.nearestObjectInTriggers.objectId);
            local pipeNode = specForageWagon.spec_pipe.nodes[1];
            if isMotorStarted and currentPipeTargetState == 2 and not trailer then
                local state = pipeNode.states[specForageWagon.spec_pipe.targetState];
                local targetRotation = state.rotation[2];
                targetRotation = MathUtil.normalizeRotationForShortestPath(targetRotation, pipeNode.curRotation[2]);
                if pipeNode.curRotation[2] < targetRotation then
                    pipeNode.curRotation[2] = math.min(pipeNode.curRotation[2] + dt * pipeNode.rotationSpeeds[2], targetRotation);
                else
                    pipeNode.curRotation[2] = math.max(pipeNode.curRotation[2] - dt * pipeNode.rotationSpeeds[2], targetRotation);
                end;
                if pipeNode.curRotation[2] > 2*math.pi then
                    pipeNode.curRotation[2] = pipeNode.curRotation[2] - 2 * math.pi;
                elseif pipeNode.curRotation[2] < -2*math.pi then
                    pipeNode.curRotation[2] = pipeNode.curRotation[2] + 2 * math.pi;
                end;
                setRotation(pipeNode.node, 0, pipeNode.curRotation[2], 0);
            end;

            for cutter, _ in pairs(controlledCombine.attachedCutters) do
                controlledCombine.cutter = cutter;
                break;
            end;

            if controlledCombine.isFilling and not controlledCombine.isSwathActive or controlledCombine.isFilling and controlledCombine.cutter.spec_cutter.strawRatio == 0 then
                g_effectManager:setFillType(specForageWagon.fillEffects, FillType.MAIZE);
                g_effectManager:startEffects(specForageWagon.fillEffects);

                if specDischargeable.dischargeSample ~= nil and not g_soundManager:getIsSamplePlaying(specDischargeable.dischargeSample) then
                    g_soundManager:playSample(specDischargeable.dischargeSample);
                end;
            else
                g_effectManager:stopEffects(specForageWagon.fillEffects);

                if specDischargeable.dischargeSample ~= nil and g_soundManager:getIsSamplePlaying(specDischargeable.dischargeSample) then
                    g_soundManager:stopSample(specDischargeable.dischargeSample);
                end;
            end;

            if self:getFillUnitFillLevelPercentage(specForageWagon.fillUnitIndex) >= 0.9 then
                self.needToStop = true;
                if actionEvent ~= nil then
                    g_inputBinding:setActionEventActive(actionEvent.actionEventId, false);
                end;

                if specForageWagon.stopTimer < specForageWagon.stopTimerMax then
                    if modCP ~= nil then
                        if vehicleToCheck:getIsCpActive() then
                            g_currentMission:showBlinkingWarning(g_i18n:getText("warning_cpIsActive"), stopTimerMax);
                        end;
                    end;
                    if not self:getIsAIActive() then
                        g_currentMission:showBlinkingWarning(g_i18n:getText("warning_almostFull"), stopTimerMax);
                        self:lockCombine(self.spec_wheels, self.spec_pipe, specForageWagon, controlledTractor);
                    end;
                    self:deactivateChaffSwath(specForageWagon.needTurnOffStrawSwath, specForageWagon, controlledTractor, vehicleToCheck, modCP);

                    specForageWagon.stopTimer = specForageWagon.stopTimer + dt;
                end;

                if self:getIsAIActive() then
                    if not specForageWagon.mapHotspotBlinkingActive then
                        self:startBeaconLights(specForageWagon.needTurnOnBeaconLights, specForageWagon, controlledTractor);
                        controlledCombine:setAIMapHotspotBlinking(true);
                        specForageWagon.mapHotspotBlinkingActive = true;
                    end;

                    if not specForageWagon.notificationGrainTankWarningShown then
                        g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_CRITICAL, string.format(g_i18n:getText("ai_messageErrorWaitUnloadChaff"), controlledCombine:getCurrentHelper().name) );
                        specForageWagon.notificationGrainTankWarningShown = true;
                    end;
                end;
            elseif self:getFillUnitFreeCapacity(specForageWagon.fillUnitIndex) == self:getFillUnitCapacity(specForageWagon.fillUnitIndex) then
                self.needToStop = false;
                if actionEvent ~= nil then
                    g_inputBinding:setActionEventActive(actionEvent.actionEventId, true);
                end;
                self:activateChaffSwath(specForageWagon.chaffSwathWasTurnedOff, specForageWagon, controlledTractor);

                if self:getIsAIActive() then
                    if specForageWagon.mapHotspotBlinkingActive then
                        self:stopBeaconLights(specForageWagon.beaconLightsWasTurnedOn, specForageWagon, controlledTractor);
                        controlledCombine:setAIMapHotspotBlinking(false);
                        specForageWagon.mapHotspotBlinkingActive = false;
                    end

                    specForageWagon.notificationGrainTankWarningShown = false;
                end;
            else
                if specForageWagon.stopTimer > 0 then
                    self:resetTimer(specForageWagon);
                end;
            end;
        end;
    end;
end;

function kskuForageExtension:resetTimer(specForageWagon)
    specForageWagon.stopTimer = 0;
end;

function kskuForageExtension:lockCombine(specWheels, specPipe, specForageWagon, controlledTractor)
    if self.isServer then
        if controlledTractor ~= nil then
            for _, wheel in pairs(controlledTractor.spec_wheels.wheels) do
                setWheelShapeProps(wheel.node, wheel.wheelShape, 0, math.huge, 0, 0);
            end;
        end;
    end;

    controlledTractor:setBrakeLightsVisibility(true);

    if self.getCruiseControlState ~= nil and self:getCruiseControlState() ~= Drivable.CRUISECONTROL_STATE_OFF then
        self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_OFF);
    end;

    if controlledTractor ~= nil and controlledTractor.getCruiseControlState ~= nil and controlledTractor:getCruiseControlState() ~= Drivable.CRUISECONTROL_STATE_OFF then
        controlledTractor:setCruiseControlState(Drivable.CRUISECONTROL_STATE_OFF);
    end;
end;

function kskuForageExtension:startBeaconLights(needTurnOnBeaconLights, specForageWagon, controlledTractor)
    if needTurnOnBeaconLights then
        controlledTractor:setBeaconLightsVisibility(true, true, nil);

        specForageWagon.needTurnOnBeaconLights = false;
        specForageWagon.beaconLightsWasTurnedOn = true;
    end;
end;

function kskuForageExtension:stopBeaconLights(beaconLightsWasTurnedOn, specForageWagon, controlledTractor)
    if beaconLightsWasTurnedOn then
        controlledTractor:setBeaconLightsVisibility(false, true, nil);

        specForageWagon.needTurnOnBeaconLights = true;
        specForageWagon.beaconLightsWasTurnedOn = false;
    end;
end;

function kskuForageExtension:deactivateChaffSwath(needTurnOffStrawSwath, specForageWagon, controlledTractor, vehicleToCheck, modCP)
    if needTurnOffStrawSwath then
        controlledTractor:setIsSwathActive(false);
        if modCP ~= nil then
            if controlledCombine.cutter ~= nil and vehicleToCheck:getIsCpActive() then
                controlledCombine.cutter.spec_cutter.strawRatio = 0;
            end;
        end;

        specForageWagon.needTurnOffStrawSwath = false;
        specForageWagon.chaffSwathWasTurnedOff = true;
    end;
end;

function kskuForageExtension:activateChaffSwath(chaffSwathWasTurnedOff, specForageWagon, controlledTractor)
    if chaffSwathWasTurnedOff then
        controlledTractor:setIsSwathActive(true);
        if controlledCombine.cutter ~= nil then
            local strawRatio = controlledCombine.cutter.spec_cutter.xmlFile:getValue ("vehicle.cutter#strawRatio", 1);
            controlledCombine.cutter.spec_cutter.strawRatio = strawRatio;
        end;

        specForageWagon.needTurnOffStrawSwath = true;
        specForageWagon.chaffSwathWasTurnedOff = false;
    end;
end;